from typing import Type

from django.conf import settings
from django.utils.datastructures import MultiValueDict
from rest_framework.authentication import BaseAuthentication, get_authorization_header, SessionAuthentication
from rest_framework.exceptions import AuthenticationFailed
from rest_framework.request import Request

from .models import BaseToken
from .settings import base_settings

try:
    import jwt
except ImportError:
    jwt = None

__all__ = ['CsrfExemptSessionAuthentication', 'TokenAuthentication']


class CsrfExemptSessionAuthentication(SessionAuthentication):
    def enforce_csrf(self, request: Request):
        return


class TokenAuthentication(BaseAuthentication):
    AUTHORIZATION_PREFIX = 'Bearer'
    AUTHORIZATION_PREFIX_BYTES = AUTHORIZATION_PREFIX.encode()

    model: Type[BaseToken] = base_settings.AUTHENTICATION_MODEL

    def __init__(self, *args, **kwargs):
        if jwt is None:
            raise ImportError(
                'PyJWT must be installed to use TokenAuthentication. Try `pip install django-rest-base[jwt]`.')
        if not isinstance(self.model, BaseToken):
            raise RuntimeError(
                f'BaseToken must be provided to either settings.REST_BASE.AUTHENTICATION_MODEL '
                f'or {self.__class__.__name__}.model'
            )

        super().__init__(*args, **kwargs)

    def authenticate(self, request: Request):
        auth = get_authorization_header(request).split()

        if not auth or auth[0] != TokenAuthentication.AUTHORIZATION_PREFIX_BYTES:
            return None

        if len(auth) != 2:
            raise AuthenticationFailed(detail='Invalid Bearer format.')

        try:
            jwt_payload = auth[1].decode()
        except UnicodeError:
            raise AuthenticationFailed(detail='Invalid Bearer encoding type.')

        header = jwt.get_unverified_header(jwt_payload)
        public_key = header.get('public_key')
        nonce = header.get('nonce')

        try:
            nonce = int(nonce)
        except ValueError:
            raise AuthenticationFailed(detail='Invalid nonce.')

        token = self.model.get(public_key, nonce)
        if token is None:
            raise AuthenticationFailed(detail='Invalid public_key or nonce.')

        try:
            payload = jwt.decode(jwt_payload, token.secret_key, algorithms='HS256')
        except jwt.exceptions.InvalidTokenError as e:
            if settings.DEBUG:
                raise AuthenticationFailed(detail=f'Invalid JWT: {e}')
            else:
                raise AuthenticationFailed(detail='Invalid JWT.')

        query = payload.get('query')
        if query:
            query_params = request.query_params
            data = request.data
            if isinstance(query_params, MultiValueDict):
                query_params = query_params.dict()
            if isinstance(data, MultiValueDict):
                data = data.dict()

            if (not data and query == query_params) or (not query_params and query == data):
                return token.user, token
        else:
            if not request.query_params and not request.data:
                return token.user, token

        raise AuthenticationFailed(detail='Query not match.')

    def authenticate_header(self, request):
        return TokenAuthentication.AUTHORIZATION_PREFIX



try:
    import numpy as np
except ImportError as e:
    raise ImportError(
        'numpy must be installed to use rest_base.utils.random. Try `pip install django-rest-base[random]`.'
    ) from e

from typing import Union

__all__ = ['seedrandom_int', 'seedrandom']

np.warnings.filterwarnings('ignore')


def _rshift(t: np.int32, n: int) -> np.int32:
    return np.int32((t % 0x100000000) >> n)


# Mulberry32
def seedrandom_int(seed: int) -> int:
    t = np.int32((seed + 0x6D2B79F5) % 2147483647)
    with np.errstate(over='ignore'):
        t = (t ^ _rshift(t, 15)) * (t | 1)
        t ^= t + (t ^ _rshift(t, 7)) * (t | 61)
    return int((t ^ _rshift(t, 14)) % 0x100000000)


def _seedrandom_str(seed: str) -> float:
    t = seedrandom_int(ord(seed[0]))
    for c in seed[1:]:
        t = seedrandom_int(t ^ seedrandom_int(ord(c)))
    return int(t) / 4294967296


def seedrandom(seed: Union[int, str]) -> float:
    if isinstance(seed, str):
        return _seedrandom_str(seed)
    return seedrandom_int(seed) / 4294967296



from django.conf import settings
from django.utils import timezone

from rest_base.utils.registry import ModuleRegistry

__all__ = ['base_settings']


DJANGO_REST_BASE_DEFAULT = {
    # commands
    'PREDEFINED_ROOT': 'predefined',

    # admin
    'ADMIN_NULL_STRING': '-',
    'MAX_FOREIGN_OBJECT_PREVIEW': 8,

    # authentication
    'AUTHENTICATION_MODEL': None,

    # errors
    'SENTRY_HOST': None,
    'SENTRY_VERBOSE': False,

    # models
    'MAX_UNIQUE_COLLISION_CHECK': 16,
    'DB_BATCH_SIZE': 1024,
    'DEFAULT_TOKEN_DURATION': timezone.timedelta(days=1),

    'USERNAME_LENGTH_MAX': 24,
    'PASSWORD_LENGTH_MAX': 128,

    'NONCE_WINDOW_SIZE': 8,
    'NONCE_MAX': 1 << 31,

    # utils
    'IP_HEADERS': ['HTTP_CF_CONNECTING_IP', 'HTTP_X_FORWARDED_FOR', 'REMOTE_ADDR'],
}

base_settings = ModuleRegistry('base_settings', include_module=True)
base_settings.update(DJANGO_REST_BASE_DEFAULT)

user_settings = getattr(settings, 'REST_BASE', None)
if user_settings is not None:
    base_settings.update(user_settings)



try:
    import channels
except ImportError as e:
    raise ImportError(
        'channels must be installed to use rest_base.consumers. Try `pip install django-rest-base[channels]`.'
    ) from e

from channels.generic.websocket import WebsocketConsumer, DenyConnection

__all__ = ['NullConsumer']


class NullConsumer(WebsocketConsumer):
    def connect(self):
        raise DenyConnection



import secrets
from abc import ABC
from inspect import isclass
from math import ceil
from typing import Type, Callable, Any

from django.db import ProgrammingError, OperationalError
from django.db.models import Model, Field

from rest_base.models import BaseModel
from rest_base.settings import base_settings
from rest_base.utils.registry import ModuleRegistry

__all__ = [
    'initialize_base_fields', 'PredefinedDefault',
    'UniqueRandom',
    'UniqueRandomInt', 'UniqueRandomPositiveInt32', 'UniqueRandomPositiveInt52', 'UniqueRandomPositiveInt64',
    'UniqueRandomChar',
]


def not_found():
    return None


unique_random = ModuleRegistry('unique_random', default=not_found)


class PredefinedDefault(ABC):
    def __init__(self):
        raise ProgrammingError(f'{self.__class__.__name__} must not be initialized')

    @classmethod
    def get_default_func(cls, model: Type[Model], field: Field) -> Callable[[], Any]:
        raise NotImplementedError('get_default_func must be defined')


class UniqueRandom(PredefinedDefault, ABC):
    pass


class UniqueRandomInt(UniqueRandom):
    """
    Unique random default int in the range [min_val, max_val)
    """

    min_val = None
    max_val = None

    @classmethod
    def get_default_func(cls, model: Type[Model], field: Field) -> Callable[[], int]:
        min_val = cls.min_val
        max_val = cls.max_val

        if min_val is None:
            raise NotImplementedError(f'{cls.__name__}.min_val must be provided')
        if max_val is None:
            raise NotImplementedError(f'{cls.__name__}.max_val must be provided')
        if type(min_val) is not int:
            raise TypeError(f'{cls.__name__}.min_val must be int but value is {min_val}')
        if type(max_val) is not int:
            raise TypeError(f'{cls.__name__}.max_val must be int but value is {max_val}')

        val_gap = max_val - min_val
        field_name = field.name
        max_collision_check = base_settings.MAX_UNIQUE_COLLISION_CHECK

        if val_gap < 1:
            raise ValueError(f'{cls.__name__}.max_val - {cls.__name__}.min_val must be bigger than 0')

        def _random():
            val = None
            try:
                for _ in range(max_collision_check):
                    val = secrets.randbelow(val_gap) - min_val
                    if not model.objects.filter(**{field_name: val}).exists():
                        return val
            except ProgrammingError:
                return val
            raise OperationalError(f'Cannot find unique value of {model.__name__}.{field_name}')

        return _random


class UniqueRandomPositiveInt(UniqueRandomInt):
    min_val = 0


class UniqueRandomPositiveInt32(UniqueRandomPositiveInt):
    max_val = 1 << 31


class UniqueRandomPositiveInt52(UniqueRandomPositiveInt):
    max_val = 1 << 51


class UniqueRandomPositiveInt64(UniqueRandomPositiveInt):
    max_val = 1 << 63


class UniqueRandomChar(UniqueRandom):
    """
    Unique random default max_len url-safe characters
    """

    @classmethod
    def get_default_func(cls, model: Type[Model], field: Field) -> Callable[[], int]:
        field_name = field.name
        length = field.max_length
        max_collision_check = base_settings.MAX_UNIQUE_COLLISION_CHECK

        if type(length) is not int:
            raise TypeError(
                f'{model.__name__}.{field_name} must have integer max_length but value is {length}')

        val_len = ceil(length * 3 / 4)

        def _random():
            val = None
            try:
                for _ in range(max_collision_check):
                    val = secrets.token_urlsafe(val_len)[:length]
                    if not model.objects.filter(**{field_name: val}).exists():
                        return val
            except ProgrammingError:
                return val
            raise OperationalError(f'Cannot find unique value of {model.__name__}.{field_name}')

        return _random


def initialize_base_fields(model: Type[BaseModel]) -> Type[BaseModel]:
    app_label = model._meta.app_label
    model_name = model._meta.model_name

    for field in model._meta._get_fields(reverse=False):
        field_name = field.name

        if field.is_relation and field.remote_field.related_name is None:
            raise ValueError(
                f"{model_name}.{field_name}.related_name must be explicitly provided."
                f"You can set '+' to omit backwards relation."
            )

        default = field.default
        if not isclass(default) or not issubclass(default, PredefinedDefault):
            continue

        default: Type[PredefinedDefault]
        if issubclass(default, UniqueRandom) and not field.unique:
            raise ValueError(f'{model_name}.{field_name}.unique must be True to use UniqueRandom')

        func_name = f'{app_label}_{model_name}_{default}'
        default_func = default.get_default_func(model, field)

        unique_random[func_name] = default_func
        field.default = default_func

    return model



try:
    import channels
except ImportError as e:
    raise ImportError(
        'channels must be installed to use rest_base.routing. Try `pip install django-rest-base[channels]`.'
    ) from e

from typing import List, Union

from channels.routing import URLRouter
from django.urls import URLPattern, URLResolver, re_path

from .consumers import NullConsumer

__all__ = ['NullURLRouter']


class NullURLRouter(URLRouter):
    def __init__(self, routes: List[Union[URLPattern, URLResolver]]):
        routes.append(re_path(r'', NullConsumer))
        super().__init__(routes)



from abc import ABC

from django.core.management import BaseCommand as DjangoBaseCommand
from django.utils import timezone

__all__ = ['BaseCommand']


class BaseCommand(DjangoBaseCommand, ABC):
    def log(self, *args):
        _args = [str(s) for s in args]
        self.stdout.write(f"[{timezone.now().strftime('%Y-%m-%d %H:%M:%S.%f')}] {' '.join(_args)}")



from typing import Callable

from rest_framework.decorators import api_view, permission_classes
from rest_framework.permissions import BasePermission
from rest_framework.request import Request
from rest_framework.response import Response
from rest_framework.settings import api_settings

__all__ = ['method_branch']


def method_branch(
        *,
        GET: Callable[..., Response] = None,
        POST: Callable[..., Response] = None,
        PUT: Callable[..., Response] = None,
        DELETE: Callable[..., Response] = None,
):
    methods = list()
    get_permissions, post_permissions, put_permissions, delete_permissions = None, None, None, None
    if GET is not None:
        methods.append('GET')
        get_permissions = getattr(GET, 'permission_classes', api_settings.DEFAULT_PERMISSION_CLASSES)
        get_permissions = [permission() for permission in get_permissions]
    if POST is not None:
        methods.append('POST')
        post_permissions = getattr(POST, 'permission_classes', api_settings.DEFAULT_PERMISSION_CLASSES)
        post_permissions = [permission() for permission in post_permissions]
    if PUT is not None:
        methods.append('PUT')
        put_permissions = getattr(PUT, 'permission_classes', api_settings.DEFAULT_PERMISSION_CLASSES)
        put_permissions = [permission() for permission in put_permissions]
    if DELETE is not None:
        methods.append('DELETE')
        delete_permissions = getattr(DELETE, 'permission_classes', api_settings.DEFAULT_PERMISSION_CLASSES)
        delete_permissions = [permission() for permission in delete_permissions]

    class PermissionClass(BasePermission):
        def has_permission(self, request, view):
            if request.method == 'GET':
                if get_permissions is None:
                    return True
                for permission in get_permissions:
                    if not permission.has_permission(request, view):
                        return False
                return True
            if request.method == 'POST':
                if post_permissions is None:
                    return True
                for permission in post_permissions:
                    if not permission.has_permission(request, view):
                        return False
                return True
            if request.method == 'PUT':
                if put_permissions is None:
                    return True
                for permission in put_permissions:
                    if not permission.has_permission(request, view):
                        return False
                return True
            if request.method == 'DELETE':
                if delete_permissions is None:
                    return True
                for permission in delete_permissions:
                    if not permission.has_permission(request, view):
                        return False
                return True
            return True

    def branch(request: Request, *args, **kwargs):
        if request.method == 'GET':
            return GET(request, *args, **kwargs)
        if request.method == 'POST':
            return POST(request, *args, **kwargs)
        if request.method == 'PUT':
            return PUT(request, *args, **kwargs)
        if request.method == 'DELETE':
            return DELETE(request, *args, **kwargs)

    return api_view(methods)(permission_classes((PermissionClass,))(branch))



from typing import Type

from django import urls
from django.contrib import admin
from django.contrib.admin import ModelAdmin
from django.contrib.admin.widgets import FilteredSelectMultiple
from django.contrib.auth import get_user_model
from django.contrib.auth.models import Group
from django.db import models
from django.db.models.fields import related
from django.forms import ModelForm, ModelMultipleChoiceField
from django.utils.html import escape
from django.utils.safestring import mark_safe

from rest_base.models import BaseModel, BaseUser
from rest_base.settings import base_settings

__all__ = ['model_admin']

User: Type[BaseUser] = get_user_model()


NULL = base_settings.ADMIN_NULL_STRING
MAX_FOREIGN_OBJECT_PREVIEW = base_settings.MAX_FOREIGN_OBJECT_PREVIEW


def _get_link_method(field: related.ForeignKey):
    field_name = field.name
    foreign_model = field.remote_field.model

    def link_method(self, instance):
        foreign_instance = getattr(instance, field_name)
        if foreign_instance is None:
            return NULL

        href = urls.reverse(
            f'admin:{foreign_model._meta.app_label.lower()}_{foreign_model._meta.object_name.lower()}_change',
            args=[foreign_instance.pk],
        )
        instance_str = escape(str(foreign_instance))

        return mark_safe(f'<a href="{href}" target="_blank">{instance_str}</a>')

    link_method.__name__ = field.name

    return link_method


def _get_multiple_link_method(field: related.ManyToManyField):
    field_name = field.name
    foreign_model = field.remote_field.model

    def link_method(self, instance):
        link_list = []
        for foreign_instance in getattr(instance, field_name).all()[:MAX_FOREIGN_OBJECT_PREVIEW]:
            href = urls.reverse(
                f'admin:{foreign_model._meta.app_label.lower()}_{foreign_model._meta.object_name.lower()}_change',
                args=[foreign_instance.pk],
            )
            instance_str = escape(str(foreign_instance))

            link_list.append(f'<a href="{href}" target="_blank">{instance_str}</a>')

        if not link_list:
            return NULL

        return mark_safe('<br>'.join(link_list))

    link_method.__name__ = field.name

    return link_method


def model_admin(model: Type[models.Model], search_fields: list = None):
    primary_key_display = []
    autocomplete_fields = []
    if search_fields is None:
        search_fields = []

    list_display = []
    list_display_sub = []
    list_display_link = dict()

    base_fields = set(f.name for f in BaseModel._meta.get_fields())

    for field in model._meta._get_fields(reverse=False):
        if not primary_key_display and field.primary_key:
            primary_key_display = [field.name]
            search_fields.append(field.name)
        elif isinstance(field, models.ForeignKey):
            autocomplete_fields.append(field.name)
            list_display.append(field.name + '_link')
            list_display_link[field.name + '_link'] = _get_link_method(field)
        elif isinstance(field, models.ManyToManyField):
            list_display.append(field.name + '_link')
            list_display_link[field.name + '_link'] = _get_multiple_link_method(field)
        elif field.name in base_fields:
            list_display_sub.append(field.name)
        else:
            list_display.append(field.name)

    return model, type(
        model.__name__ + 'Admin',
        (ModelAdmin,),
        dict(
            search_fields=search_fields,
            autocomplete_fields=autocomplete_fields,

            list_display=primary_key_display + list_display + list_display_sub,
            **list_display_link
        )
    )


class GroupAdminForm(ModelForm):
    class Meta:
        model = Group
        exclude = []

    users = ModelMultipleChoiceField(
         queryset=User.objects.all(),
         required=False,
         widget=FilteredSelectMultiple('users', False)
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self.instance.pk:
            self.fields['users'].initial = self.instance.user_set.all()

    def _save_m2m(self):
        super()._save_m2m()
        self.instance.user_set.set(self.cleaned_data['users'])


class GroupAdmin(ModelAdmin):
    form = GroupAdminForm
    list_display = ['name', 'permissions_list', 'users']
    filter_horizontal = ['permissions']

    search_fields = ['name']

    def permissions_list(self, instance):
        link_list = []
        for foreign_instance in instance.permissions.all()[:MAX_FOREIGN_OBJECT_PREVIEW]:
            link_list.append(escape(str(foreign_instance)))
        if not link_list:
            return NULL
        return mark_safe('<br>'.join(link_list))
    permissions_list.__name__ = 'permissions'

    def users(self, instance):
        link_list = []
        for user in instance.user_set.all()[:MAX_FOREIGN_OBJECT_PREVIEW]:
            href = urls.reverse(
                f'admin:{User._meta.app_label.lower()}_{User._meta.object_name.lower()}_change',
                args=[user.pk],
            )
            user_str = escape(str(user))

            link_list.append(f'<a href="{href}" target="_blank">{user_str}</a>')

        if not link_list:
            return NULL
        return mark_safe('<br>'.join(link_list))


admin.site.unregister(Group)
admin.site.register(Group, GroupAdmin)



from __future__ import annotations

import traceback
from typing import TypedDict, Literal, Optional, Union, Any

from django.conf import settings
from django.core.exceptions import PermissionDenied
from django.http import HttpRequest, Http404
from rest_framework import status
from rest_framework.exceptions import APIException
from rest_framework.request import Request
from rest_framework.response import Response
from rest_framework.views import APIView
from rest_framework.views import exception_handler as rest_exception_handler

from rest_base.settings import base_settings

__all__ = [
    'SENTRY_ERROR_LEVEL', 'CODE_UNKNOWN',
    'Error', 'sentry_report', 'ExceptionHandlerContext', 'exception_handler',
]

SENTRY_ERROR_LEVEL = Literal['debug', 'info', 'warning', 'error', 'fatal']
CODE_UNKNOWN = 'Unknown'


sentry_enabled = bool(getattr(base_settings, 'SENTRY_HOST', False))
sentry_verbose = getattr(base_settings, 'SENTRY_VERBOSE', False)

if sentry_enabled:
    try:
        import sentry_sdk
        from sentry_sdk.integrations.django import DjangoIntegration

        sentry_sdk.init(base_settings.SENTRY_HOST, integrations=[sentry_sdk.integrations.django.DjangoIntegration()])
    except ImportError:
        sentry_sdk = None
        DjangoIntegration = None
else:
    sentry_sdk = None
    DjangoIntegration = None


class Error(APIException):
    status_code = status.HTTP_400_BAD_REQUEST

    class SerializedCode(TypedDict):
        code: str

    class SerializedDetail(SerializedCode, total=False):
        detail: str
        extra: Any
        traceback: str

    class Serialized(TypedDict):
        error: Error.SerializedDetail

    def __init__(
            self, app_or_request: Union[str, HttpRequest, Request],
            *args: str, code: str = None, detail: str = None, extra: Any = None,
            tb: str = None, status_code: int = None,
    ):
        if type(app_or_request) is str:
            self.app = app_or_request
        elif isinstance(app_or_request, HttpRequest):
            self.app = app_or_request.resolver_match.app_name
        elif isinstance(app_or_request, Request):
            self.app = app_or_request.resolver_match.app_name
        else:
            raise ValueError(
                f'app_or_request must be str, HttpRequest, or Request but {type(app_or_request)} provided')

        self.code = code or '::'.join(args) or CODE_UNKNOWN
        self.serialized: Error.Serialized = Error.Serialized(
            error=dict(
                code=self.code,
            ),
        )
        if detail is not None:
            self.serialized['error']['detail'] = detail
        if extra is not None:
            self.serialized['error']['extra'] = extra
        if tb is not None:
            self.serialized['error']['traceback'] = tb
        if status_code is not None:
            self.status_code = status_code

        super().__init__(self.serialized, self.code)

    def __str__(self):
        return f'Error ({self.app}::{self.code})'

    def __call__(self, *args, code: str = None, detail: str = None, extra: Any = None, tb: str = None) -> Error:
        code = code or '::'.join(args) or self.code
        return Error(self.app, code=code, detail=detail, extra=extra, tb=tb)


def sentry_report(
        exc: Exception, level: SENTRY_ERROR_LEVEL = 'error', silent: bool = True
) -> Optional[int]:
    global sentry_enabled, sentry_verbose

    if not sentry_enabled:
        if not silent:
            raise RuntimeError('SENTRY_HOST is not found from django.conf.settings')
        return

    if sentry_sdk is None:
        if not silent:
            raise ImportError('Module sentry_sdk is not found. Try `pip install django-rest-base[sentry]`.')
        return

    if level in ('debug', 'info') and not sentry_verbose:
        return

    with sentry_sdk.configure_scope() as scope:
        scope.set_level(level)
        try:
            return sentry_sdk.capture_exception(exc)
        except Exception as e:
            print('sentry error:', e)
            print(traceback.format_exc())
            if not silent:
                raise


class ExceptionHandlerContext(TypedDict):
    view: APIView
    args: tuple
    kwargs: dict
    request: Request


def exception_handler(exc: Exception, context: ExceptionHandlerContext) -> Optional[Response]:
    if isinstance(exc, Error):
        sentry_report(exc, level='debug')

        if settings.DEBUG:
            if context['request'].content_type.startswith('application/json'):
                exc = exc(traceback=traceback.format_exc())
            else:
                return None
    elif isinstance(exc, APIException):
        sentry_report(exc, level='debug')
        exc = Error(context['request'], exc.__class__.__name__, detail=exc.detail, status_code=exc.status_code)
    elif not isinstance(exc, (Http404, PermissionDenied)):
        event_id = sentry_report(exc)
        event_id = f'event_id: {event_id}' if event_id is not None else None

        if settings.DEBUG:
            return None
        else:
            exc = Error(context['request'], detail=event_id)

    return rest_exception_handler(exc, context)



from __future__ import annotations

import importlib
import re
from typing import Union, Any

__all__ = ['ModuleRegistry']


MODULE_NAME_REGEX = re.compile(r'^(?:[0-9a-zA-Z_]+\.)+[0-9a-zA-Z_]+$')


def _try_import(val: str):
    if MODULE_NAME_REGEX.match(val) is None:
        return val

    path, name = val.rsplit('.', 1)
    try:
        module = importlib.import_module(path)
    except ImportError:
        return val
    return getattr(module, name, val)


class EMPTY:
    pass


class INVALID:
    pass


class ModuleRegistry:
    def __init__(self, name: str, default: Any = EMPTY, include_module: bool = False):
        super().__setattr__('name', name)
        super().__setattr__('_registry', dict())
        super().__setattr__('_default', default)
        super().__setattr__('_include_module', include_module)

    def __getattr__(self, name, default=INVALID):
        value = self._registry.get(name, default)
        if value is INVALID:
            if self._default is EMPTY:
                raise AttributeError(f'{self.name} has no attribute {name}')
            else:
                return None
        return value

    def __setattr__(self, name, value):
        self._set(name, value)

    def __delattr__(self, name):
        try:
            del self._registry[name]
        except KeyError:
            if not self._ignore:
                raise AttributeError(f'{self.name} has no attribute {name}')

    def __getitem__(self, key):
        try:
            return self._registry[key]
        except KeyError:
            if self._ignore:
                return None
            raise

    def __setitem__(self, key, value):
        self._set(key, value)

    def __delitem__(self, key):
        try:
            del self._registry[key]
        except KeyError:
            if self._ignore:
                return None
            raise

    def _set(self, key, value):
        if callable(value):
            value.__name__ = key
            value.__qualname__ = f'{self.name}.{key}'
        elif isinstance(value, str):
            value = _try_import(value)

        self._registry[key] = value

    def dict(self) -> dict:
        return self._registry

    def update(self, other: Union[dict, ModuleRegistry]):
        if isinstance(other, dict):
            self._registry.update(other)
        elif isinstance(other, ModuleRegistry):
            self._registry.update(other._registry)
        else:
            raise TypeError('Only dictionaries and ModuleRegistry are allowed')



import setuptools

with open('README.md', 'r', encoding='utf8') as f:
    long_description = f.read()

setuptools.setup(
    name='django-rest-base',
    version='0.1.5',
    url='https://github.com/devluci/django-rest-base',
    author='Lucid (@devluci)',
    author_email='contact@lucid.dev',
    description='Customized features and environment for building a Django REST framework app.',
    license='MIT',
    long_description=long_description,
    long_description_content_type='text/markdown',
    packages=setuptools.find_packages(),
    include_package_data=True,
    python_requires='>=3.8',
    install_requires=[
        'Django==3.*',
        'djangorestframework>=3.10',
    ],
    extras_require={
        'jwt': ['PyJWT'],
        'channels': ['channels'],
        'sentry': ['sentry-sdk'],
        'random': ['numpy'],
    },
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Web Environment',
        'Topic :: Internet :: WWW/HTTP',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.8',
        'Framework :: Django',
        'Framework :: Django :: 3.0',
    ],
)



import os

from django.core.management.commands.startapp import Command as StartAppCommand

import rest_base


class Command(StartAppCommand):
    def handle(self, **options):
        if options['template'] is None:
            options['template'] = os.path.join(rest_base.__path__[0], 'conf', 'app_template')

        super().handle(**options)
